Imports ExpressionLib

Imports System.CodeDom.Compiler
Imports System.Text
Imports System.Reflection
Imports System.Math
Imports System.Collections.Generic

Public Class Form1
    Private m_NumTests As Integer
    Private m_NumPassed As Integer

    Private m_Random As New Random()

    Private Sub btnRunTests_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles btnRunTests.Click
        Debug.WriteLine("")
        Debug.WriteLine("")
        Debug.WriteLine("")

        lvwTests.Items.Clear()
        m_NumTests = 0
        m_NumPassed = 0

        ' Make some parameters.
        Dim variables As New Dictionary(Of String, Double)
        variables.Add("A", 123.456)
        variables.Add("B", 7463.218)
        variables.Add("C", 0.000000367)
        variables.Add("D", 2.235E+210)

        ' Simple operator tests.
        lvwTests.Items.Add("Simple Tests").BackColor = Color.PaleGreen
        RunTest(".1", variables)
        RunTest("0.1", variables)
        RunTest("1234", variables)
        RunTest("A", variables)
        RunTest("(17)", variables)
        RunTest("-13", variables)
        RunTest("3 * -2", variables)
        RunTest("3*-2", variables)
        RunTest("+37", variables)
        RunTest("2 * +3", variables)
        RunTest("2*+3", variables)
        RunTest("3^2", variables)
        RunTest("6*7", variables)
        RunTest("4/5", variables)
        RunTest("4\5", variables)
        RunTest("7%4", variables)
        RunTest("3+9", variables)
        RunTest("3-9", variables)
        RunTest("Sin(A)", variables)
        RunTest("Cos(B)", variables)
        RunTest("Tan(C)", variables)
        RunTest("Sqrt(D)", variables)
        RunTest("Log10(1010101)", variables)
        RunTest("Log(1010101)", variables)
        RunTest("Exp(13.37)", variables)
        RunTest("ATan(1)", variables)
        RunTest("ATan(-21)", variables)
        RunTest("ATan(12)", variables)
        RunTest("ASin(.5)", variables)
        RunTest("ACos(.5)", variables)
        RunTest("ASin(0.5)", variables)
        RunTest("ACos(0.5)", variables)

        ' Misc tests.
        lvwTests.Items.Add("Misc Tests").BackColor = Color.PaleGreen
        RunTest("1+2", variables)
        RunTest("1+2+3", variables)
        RunTest("2+3*4", variables)
        RunTest("2+(3*4)", variables)
        RunTest("2*3+4", variables)
        RunTest("2*(3+4)", variables)
        RunTest("13^7-12/19*25", variables)
        RunTest("563 % 12", variables)
        RunTest("2*3+4", variables)
        RunTest("2*(3+4)", variables)
        RunTest("2*3+4", variables)
        RunTest("2*(3+4)", variables)
        RunTest("A*B-C/D*A", variables)

        ' Added to catch bug in Tan(-(A)^-ATan(D)).
        RunTest("-1^-2", variables)
        RunTest("(-123.456)^(-1.5707963267949)", variables)
        RunTest("-(A)^-ATan(D)", variables)
        RunTest("-(A)", variables)
        RunTest("-ATan(D)", variables)
        RunTest("Tan(-(A)^-ATan(D))", variables)

        ' Weird cases.
        lvwTests.Items.Add("Weird Cases").BackColor = Color.PaleGreen
        RunTest("2\0", variables)
        RunTest("2/0", variables)
        RunTest("1^0", variables)
        RunTest("1^-1", variables)
        RunTest("70.7^-12.3", variables)
        RunTest("Sqrt(0)", variables)
        RunTest("Sqrt(-1)", variables)
        RunTest("Exp(1337)", variables)
        RunTest("ATan(0)", variables)
        RunTest("ASin(12)", variables)
        RunTest("ACos(12)", variables)
        RunTest("", variables)
        RunTest("()", variables)
        RunTest("(-1)", variables)
        RunTest("(((-1)))", variables)
        RunTest("(-1", variables)
        RunTest("(-1))", variables)
        RunTest("-1)", variables)
        RunTest("((-1)", variables)
        RunTest("()-1", variables)

        ' Random tests.
        Dim max_expression_depth As Integer = Integer.Parse(txtMaxExpressionDepth.Text)
        lvwTests.Items.Add("Random Tests").BackColor = Color.PaleGreen
        For i As Integer = 1 To Integer.Parse(txtNumRandomTests.Text)
            ' Make a random test.
            Dim expression_string As String = RandomExpression(m_Random.Next(1, max_expression_depth))
            RunTest(expression_string, variables)
        Next i

        MessageBox.Show( _
            m_NumTests.ToString() & " tests" & vbCrLf & _
            m_NumPassed.ToString() & " passed" & vbCrLf & _
            (m_NumTests - m_NumPassed).ToString() & " failed", _
            "Done", MessageBoxButtons.OK, MessageBoxIcon.Information)
    End Sub

    ' Recursively make a random expression.
    Private Function RandomExpression(ByVal depth As Integer) As String
        If depth < 1 Then
            ' Use a constant or variable.
            Select Case m_Random.Next(8)
                Case 0 To 3
                    Return (m_Random.NextDouble * 10000).ToString()
                Case 4
                    Return "A"
                Case 5
                    Return "B"
                Case 6
                    Return "C"
                Case 7
                    Return "D"
            End Select
        Else
            ' Pick an operator.
            Select Case m_Random.Next(20)
                Case 0  ' Parens.
                    Return "(" & RandomExpression(depth - 1) & ")"
                Case 1  ' Unary minus.
                    Return "-" & RandomExpression(depth - 1)
                Case 2  ' Unary plus.
                    Return "+" & RandomExpression(depth - 1)
                Case 3  ' ^
                    Return RandomExpression(depth - 1) & " ^ " & RandomExpression(depth - 1)
                Case 4  ' *
                    Return RandomExpression(depth - 1) & " * " & RandomExpression(depth - 1)
                Case 5  ' /
                    Return RandomExpression(depth - 1) & " / " & RandomExpression(depth - 1)
                Case 6  ' \
                    Return RandomExpression(depth - 1) & " \ " & RandomExpression(depth - 1)
                Case 7  ' %
                    Return RandomExpression(depth - 1) & " % " & RandomExpression(depth - 1)
                Case 8  ' +
                    Return RandomExpression(depth - 1) & " + " & RandomExpression(depth - 1)
                Case 9  ' -
                    Return RandomExpression(depth - 1) & " - " & RandomExpression(depth - 1)
                Case 10 ' Sin
                    Return "Sin(" & RandomExpression(depth - 1) & ")"
                Case 11 ' Cos
                    Return "Cos(" & RandomExpression(depth - 1) & ")"
                Case 12 ' Tan
                    Return "Tan(" & RandomExpression(depth - 1) & ")"
                Case 13 ' Sqrt
                    Return "Sqrt(" & RandomExpression(depth - 1) & ")"
                Case 14 ' Log10
                    Return "Log10(" & RandomExpression(depth - 1) & ")"
                Case 15 ' Log
                    Return "Log(" & RandomExpression(depth - 1) & ")"
                Case 16 ' Exp
                    Return "Exp(" & RandomExpression(depth - 1) & ")"
                Case 17 ' ATan
                    Return "ATan(" & RandomExpression(depth - 1) & ")"
                Case 18 ' ASin
                    Return "ASin(" & RandomExpression(depth - 1) & ")"
                Case 19 ' ACos
                    Return "ACos(" & RandomExpression(depth - 1) & ")"
            End Select
        End If

        Return Nothing
    End Function

    ' Evaluate the expression using a CompiledExpression
    ' and compiled VB code and compare the results.
    Private Sub RunTest(ByVal expression_string As String, ByVal variables As Dictionary(Of String, Double))
        m_NumTests += 1

        Dim result1 As Double
        Dim result2 As Double
        Dim err1 As Boolean = False
        Dim err2 As Boolean = False

        ' Calculate the result using a CompiledExpression.
        Try
            Dim expr As New CompiledExpression(expression_string)
            result1 = expr.Evaluate(variables)
        Catch ex As Exception
            err1 = True
            Debug.WriteLine( _
                "*** CompiledExpression Error ***" & _
                vbCrLf & ex.Message & vbCrLf)
        End Try

        ' Calculate the result using compiled VB code.
        Try
            result2 = VbResult(expression_string.Replace("%", " Mod "), variables)
        Catch ex As Exception
            err2 = True
            Debug.WriteLine( _
                "*** CompiledExpression VB ***" & _
                vbCrLf & ex.Message & vbCrLf)
        End Try

        ' Compare the results.
        Dim diff As Double = result2 - result1

        ' Display the result.
        Dim list_view_item As ListViewItem = lvwTests.Items.Add(expression_string)

        ' Pass/Fail.
        If (err1 AndAlso err2) OrElse _
            (Double.IsNaN(result1) AndAlso Double.IsNaN(result2)) OrElse _
            (Double.IsInfinity(result1) AndAlso Double.IsInfinity(result2)) OrElse _
            (Abs(diff) < 0.000000001) _
        Then
            ' Passed.
            list_view_item.SubItems.Add("-")
            m_NumPassed += 1
        Else
            ' Failed.
            list_view_item.SubItems.Add("Fail")
            list_view_item.BackColor = Color.HotPink
        End If

        ' CompiledExpression value.
        If err1 Then
            list_view_item.SubItems.Add("Error")
        Else
            list_view_item.SubItems.Add(result1.ToString())
        End If

        ' VB value.
        If err2 Then
            list_view_item.SubItems.Add("Error")
        Else
            list_view_item.SubItems.Add(result2.ToString())
        End If

        ' Difference.
        If Not err1 AndAlso Not err2 Then
            list_view_item.SubItems.Add(diff.ToString())
        End If

        lvwTests.EnsureVisible(lvwTests.Items.Count - 1)
        lvwTests.Refresh()
    End Sub

    ' Compile some VB code and use it to perform the calculation.
    Private Function VbResult(ByVal expression_string As String, ByVal variables As Dictionary(Of String, Double)) As Double
        ' Compose the function's parameter list.
        Dim params As String = ""
        Dim param_name As String = "A"
        For i As Integer = 0 To variables.Count - 1
            params &= ", ByVal " & param_name & " As Double"
            param_name = Chr(Asc(param_name) + 1)
        Next i
        If params.Length > 0 Then params = params.Substring(2)

        ' Compose the code.
        Dim vb_code As String = _
            "Imports System" & vbCrLf & _
            "Imports System.Math" & vbCrLf & _
            "Namespace MainNamespace" & vbCrLf & _
            "  Class MainClass" & vbCrLf & _
            "    Public Function Main(" & params & ") As Double" & vbCrLf & _
            "        Return " & expression_string & vbCrLf & _
            "    End Function" & vbCrLf & _
            "  End Class" & vbCrLf & _
            "End Namespace"

        ' Prepare the compiler parameters.
        ' Add any needed references.
        Dim compiler_params As CompilerParameters = New CompilerParameters
        compiler_params.ReferencedAssemblies.Add("system.dll")

        ' Generate a library in memory.
        compiler_params.CompilerOptions = "/t:library"
        compiler_params.GenerateInMemory = True

        ' Compile the code.
        Dim code_provider As VBCodeProvider = New VBCodeProvider
        Dim compiler_results As CompilerResults = _
            code_provider.CompileAssemblyFromSource(compiler_params, vb_code)
        code_provider.Dispose()

        ' Check for errors.
        ' CompilerErrors = compiler_results.Errors
        If compiler_results.Errors.Count > 0 Then
            ' There were errors.
            Dim txt As String = "Compiler error:" & vbCrLf & vbCrLf
            For Each compiler_error As CompilerError In compiler_results.Errors
                txt &= vbTab & compiler_error.ErrorText & vbCrLf
            Next compiler_error
            Throw New Exception(txt)
        Else
            ' There were no errors. Try to execute the code.
            ' Get the assembly.
            Dim the_assembly As System.Reflection.Assembly = _
                compiler_results.CompiledAssembly

            ' Make a MainClass object.
            Dim main_object As Object = _
                the_assembly.CreateInstance("MainNamespace.MainClass")

            ' Get the MainClass's type.
            Dim main_object_type As Type = main_object.GetType

            ' Get the MainClass's Main method.
            Dim run_method_info As MethodInfo
            run_method_info = main_object_type.GetMethod("Main")

            ' Make the parameters array.
            Dim parameters(0 To variables.Count - 1) As Object
            Dim vals As Dictionary(Of String, Double).ValueCollection
            vals = variables.Values
            Dim param_num As Integer = 0
            For Each val As Double In vals
                parameters(param_num) = val
                param_num += 1
            Next val

            ' Evaluate the expression.
            Dim result As Double
            Try
                result = CDbl(run_method_info.Invoke(main_object, parameters))
            Catch ex As Exception
                ' Rethrow the exception.
                Throw ex
            End Try
            Return result
        End If
    End Function

End Class
